{% if '-g' in version %}
{% set sha = version.rsplit('-g', 1)[-1] %}
<p>These docs are for Salt's development version:
    <a href="https://github.com/saltstack/salt/tree/{{ sha }}">{{ sha }}</a>.</p>
{% endif %}

<p class="muted">Docs for previous releases are available on
    <a href="http://salt.readthedocs.org">salt.rtfd.org</a>.</p>

<p>Latest Salt release: <a href="{{ pathto('topics/releases/{0}'.format(release)) }}">{{ release }}</a></p>

<p>Try the shiny new release candidate of Salt,
    <a href="{{ pathto('topics/releases/2014.7.0') }}">v2014.7.0rc3</a>!  More info
    <a href="{{ pathto('topics/releases/releasecandidate') }}">here.</a></p>
