# This file was auto-generated by salt's setup on Thursday, 16 October 2014 @ 22:10:18 UTC.

from salt.version import SaltStackVersion

__saltstack_version__ = SaltStackVersion(2014, 7, 0, 0, 4, 0, None)
